@@ -4,7 +4,6 @@ import tempfile
 from abc import abstractmethod, ABCMeta
 
 from mlflow.utils.validation import path_not_unique, bad_path_message
-from mlflow.utils.annotations import experimental
 
 from mlflow.exceptions import MlflowException
 from mlflow.protos.databricks_pb2 import INVALID_PARAMETER_VALUE, RESOURCE_DOES_NOT_EXIST
@@ -62,6 +61,34 @@ class ArtifactRepository:
         listing = self.list_artifacts(artifact_path)
         return len(listing) > 0
 
+    def _create_download_destination(self, src_artifact_path, dst_local_dir_path=None):
+        """
+        Creates a local filesystem location to be used as a destination for downloading the artifact
+        specified by `src_artifact_path`. The destination location is a subdirectory of the
+        specified `dst_local_dir_path`, which is determined according to the structure of
+        `src_artifact_path`. For example, if `src_artifact_path` is `dir1/file1.txt`, then the
+        resulting destination path is `<dst_local_dir_path>/dir1/file1.txt`. Local directories are
+        created for the resulting destination location if they do not exist.
+
+        :param src_artifact_path: A relative, POSIX-style path referring to an artifact stored
+                                  within the repository's artifact root location.
+                                  `src_artifact_path` should be specified relative to the
+                                  repository's artifact root location.
+        :param dst_local_dir_path: The absolute path to a local filesystem directory in which the
+                                   local destination path will be contained. The local destination
+                                   path may be contained in a subdirectory of `dst_root_dir` if
+                                   `src_artifact_path` contains subdirectories.
+        :return: The absolute path to a local filesystem location to be used as a destination
+                 for downloading the artifact specified by `src_artifact_path`.
+        """
+        src_artifact_path = src_artifact_path.rstrip("/")  # Ensure correct dirname for trailing '/'
+        dirpath = posixpath.dirname(src_artifact_path)
+        local_dir_path = os.path.join(dst_local_dir_path, dirpath)
+        local_file_path = os.path.join(dst_local_dir_path, src_artifact_path)
+        if not os.path.exists(local_dir_path):
+            os.makedirs(local_dir_path, exist_ok=True)
+        return local_file_path
+
     def download_artifacts(self, artifact_path, dst_path=None):
         """
         Download an artifact file or directory to a local directory if applicable, and return a
@@ -80,32 +107,51 @@ class ArtifactRepository:
 
         # TODO: Probably need to add a more efficient method to stream just a single artifact
         #       without downloading it, or to get a pre-signed URL for cloud storage.
-        def download_file(fullpath):
-            fullpath = fullpath.rstrip("/")  # Prevents incorrect split if fullpath ends with a '/'
-            dirpath, _ = posixpath.split(fullpath)
-            local_dir_path = os.path.join(dst_path, dirpath)
-            local_file_path = os.path.join(dst_path, fullpath)
-            if not os.path.exists(local_dir_path):
-                os.makedirs(local_dir_path)
-            self._download_file(remote_file_path=fullpath, local_path=local_file_path)
-            return local_file_path
+        def download_artifact(src_artifact_path, dst_local_dir_path):
+            """
+            Download the file artifact specified by `src_artifact_path` to the local filesystem
+            directory specified by `dst_local_dir_path`.
 
-        def download_artifact_dir(dir_path):
-            local_dir = os.path.join(dst_path, dir_path)
+            :param src_artifact_path: A relative, POSIX-style path referring to a file artifact
+                                      stored within the repository's artifact root location.
+                                      `src_artifact_path` should be specified relative to the
+                                      repository's artifact root location.
+            :param dst_local_dir_path: Absolute path of the local filesystem destination directory
+                                       to which to download the specified artifact. The downloaded
+                                       artifact may be written to a subdirectory of
+                                       `dst_local_dir_path` if `src_artifact_path` contains
+                                       subdirectories.
+            :return: A local filesystem path referring to the downloaded file.
+            """
+            local_destination_file_path = self._create_download_destination(
+                src_artifact_path=src_artifact_path, dst_local_dir_path=dst_local_dir_path
+            )
+            self._download_file(
+                remote_file_path=src_artifact_path, local_path=local_destination_file_path
+            )
+            return local_destination_file_path
+
+        def download_artifact_dir(src_artifact_dir_path, dst_local_dir_path):
+            local_dir = os.path.join(dst_local_dir_path, src_artifact_dir_path)
             dir_content = [  # prevent infinite loop, sometimes the dir is recursively included
                 file_info
-                for file_info in self.list_artifacts(dir_path)
-                if file_info.path != "." and file_info.path != dir_path
+                for file_info in self.list_artifacts(src_artifact_dir_path)
+                if file_info.path != "." and file_info.path != src_artifact_dir_path
             ]
             if not dir_content:  # empty dir
                 if not os.path.exists(local_dir):
-                    os.makedirs(local_dir)
+                    os.makedirs(local_dir, exist_ok=True)
             else:
                 for file_info in dir_content:
                     if file_info.is_dir:
-                        download_artifact_dir(dir_path=file_info.path)
+                        download_artifact_dir(
+                            src_artifact_dir_path=file_info.path,
+                            dst_local_dir_path=dst_local_dir_path,
+                        )
                     else:
-                        download_file(file_info.path)
+                        download_artifact(
+                            src_artifact_path=file_info.path, dst_local_dir_path=dst_local_dir_path
+                        )
             return local_dir
 
         if dst_path is None:
@@ -131,9 +177,11 @@ class ArtifactRepository:
 
         # Check if the artifacts points to a directory
         if self._is_directory(artifact_path):
-            return download_artifact_dir(artifact_path)
+            return download_artifact_dir(
+                src_artifact_dir_path=artifact_path, dst_local_dir_path=dst_path
+            )
         else:
-            return download_file(artifact_path)
+            return download_artifact(src_artifact_path=artifact_path, dst_local_dir_path=dst_path)
 
     @abstractmethod
     def _download_file(self, remote_file_path, local_path):
@@ -147,7 +195,6 @@ class ArtifactRepository:
         """
         pass
 
-    @experimental
     def delete_artifacts(self, artifact_path=None):
         """
         Delete the artifacts at the specified location.
