@@ -8,9 +8,10 @@ import sys
 import tarfile
 import tempfile
 
-from six.moves.urllib.request import pathname2url
-from six.moves.urllib.parse import unquote
-from six.moves import urllib
+import urllib.parse
+import urllib.request
+from urllib.parse import unquote
+from urllib.request import pathname2url
 
 import yaml
 
@@ -21,6 +22,7 @@ except ImportError:
 
 from mlflow.entities import FileInfo
 from mlflow.exceptions import MissingConfigException
+from mlflow.utils.rest_utils import cloud_storage_http_request
 
 ENCODING = "utf-8"
 
@@ -122,7 +124,7 @@ def make_containing_dirs(path):
         os.makedirs(dir_name)
 
 
-def write_yaml(root, file_name, data, overwrite=False):
+def write_yaml(root, file_name, data, overwrite=False, sort_keys=True):
     """
     Write dictionary data in yaml format.
 
@@ -143,7 +145,12 @@ def write_yaml(root, file_name, data, overwrite=False):
     try:
         with codecs.open(yaml_file_name, mode="w", encoding=ENCODING) as yaml_file:
             yaml.dump(
-                data, yaml_file, default_flow_style=False, allow_unicode=True, Dumper=YamlSafeDumper
+                data,
+                yaml_file,
+                default_flow_style=False,
+                allow_unicode=True,
+                sort_keys=sort_keys,
+                Dumper=YamlSafeDumper,
             )
     except Exception as e:
         raise e
@@ -412,3 +419,20 @@ def yield_file_in_chunks(file, chunk_size=100000000):
                 yield chunk
             else:
                 break
+
+
+def download_file_using_http_uri(http_uri, download_path, chunk_size=100000000):
+    """
+    Downloads a file specified using the `http_uri` to a local `download_path`. This function
+    uses a `chunk_size` to ensure an OOM error is not raised a large file is downloaded.
+
+    Note : This function is meant to download files using presigned urls from various cloud
+            providers.
+    """
+    with cloud_storage_http_request("get", http_uri, stream=True) as response:
+        response.raise_for_status()
+        with open(download_path, "wb") as output_file:
+            for chunk in response.iter_content(chunk_size=chunk_size):
+                if not chunk:
+                    break
+                output_file.write(chunk)
